package controller;

import java.rmi.NotBoundException;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.ArrayList;
import java.util.List;
import java.util.Observable;

import javax.persistence.metamodel.ManagedType;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import main.ResultContent;

import view.AbstractDialog;
import view.AbstractFrame;
import view.LoginFrame;

/**
 * Controller des MVC-Patterns
 * 
 * @author Thomas Kramer, eMail: th-rzv@gmx.de
 * 
 */
public class Controller extends Observable {
	private Controller controller = null;
	public static ServerInterface rmiServer = null;
	public static ClientInterfaceImpl rmiClient = null;

	/**
	 * interne Liste von Frames/Dialogs
	 */
	private static ArrayList<AbstractFrame> viewFrames = new ArrayList<AbstractFrame>();
	private static ArrayList<AbstractDialog> viewDialogs = new ArrayList<AbstractDialog>();

	/**
	 * Anzahl der Frames/Dialogs die in den Arraylisten gleichzeitig vorgehalten
	 * werden
	 */
	private static int framesLimit = 2;
	private static int dialogsLimit = 2;

	public Controller(String[] args) {
		controller = this;
		boolean ok = false;

		/**
		 * Serververbindung starten
		 */
		String serverPort = String.valueOf(Registry.REGISTRY_PORT);
		String serverAddress = "192.168.178.45";

		System.setProperty("java.security.policy", "file:./client.policy");

		if (args.length > 1) {
			serverPort = args[0];
			serverAddress = args[1];
		}

		/**
		 * Erstellen des Security-Managers fr RMI-Kommunikation
		 */
		if (System.getSecurityManager() == null) {
			System.setSecurityManager(new RMISecurityManager());
		}

		try {
			Registry registry = LocateRegistry.getRegistry(serverAddress,
					(new Integer(serverPort)).intValue());
			rmiServer = (ServerInterface) (registry
					.lookup(ServerInterface.SERVICE_NAME));
			ok = true;
		} catch (java.rmi.ConnectException ce) {
			JOptionPane
					.showMessageDialog(
							null,
							"Die Server-Anwendung der Projektverwaltung wurde noch nicht gestartet!",
							"Fehler", JOptionPane.ERROR_MESSAGE);
			System.err.println("Server-Anwendung wurde noch nicht gestartet!");
		} catch (RemoteException e) {
			System.err.println(e);
		} catch (NotBoundException e) {
			System.err.println(e);
		}

		/**
		 * GUI starten
		 */
		if (ok) {
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					try {
						LoginFrame frame = new LoginFrame(controller);
					} catch (RemoteException e) {
						System.err
								.println("Fehler beim Anmelden zum Server bzw. hinzufgen zur Observer-Liste!");
						System.err.println(e);
						System.err
								.println("Anwendung wird fehlerbedingt beendet!");
						System.exit(1);
					}

					/**
					 * Das LoginFrame selbst mu nicht als Beobachter
					 * hinzugefgt werden
					 */
				}
			});
		} else {
			System.err.println("Anwendung wird fehlerbedingt beendet!");
			System.exit(1);
		}
	}

	/**
	 * Objekte vom Server anwendungsintern an alle Views ber das
	 * Observer-Pattern weiterreichen
	 * 
	 * @param object
	 *            Daten-Objekt
	 */
	public void delegateObject(Class t, List<Object> object) {
		tellChangedData(t, object);
	}

	/**
	 * einen bestimmten JFrame aus der Controller-Liste auslesen und zurckgeben
	 * 
	 * @param jFrame
	 * @return
	 */
	public AbstractFrame getFrame(Class jFrame) {
		AbstractFrame result = null;
		for (AbstractFrame view : viewFrames) {
			if (view.getClass().equals(jFrame)) {
				result = view;
				break;
			}
		}
		return result;
	}

	/**
	 * fgt einen bestimmten JFrame der internen Controller-Liste hinzu
	 * 
	 * @param jDialog
	 */
	public void addFrame(AbstractFrame jFrame) {
		viewFrames.add(jFrame);

		if (viewFrames.size() > framesLimit) {
			viewFrames.remove(0);
			System.out
					.println("Listenkapazitt der JFrames im Controller berschritten, das erste Element wurde wieder entfernt!");
		}
	}

	/**
	 * einen bestimmten JDialog aus der Controller-Liste auslesen und
	 * zurckgeben
	 * 
	 * @param jFrame
	 * @return
	 */
	public AbstractDialog getDialog(Class jDialog) {
		AbstractDialog result = null;
		for (AbstractDialog view : viewDialogs) {
			if (view.getClass().equals(jDialog)) {
				result = view;
				break;
			}
		}
		return result;
	}

	/**
	 * fgt einen bestimmten JDialog der internen Controller-Liste hinzu
	 * 
	 * @param jDialog
	 */
	public void addDialog(AbstractDialog jDialog) {
		viewDialogs.add(jDialog);

		if (viewDialogs.size() > dialogsLimit) {
			viewDialogs.remove(0);
			System.out
					.println("Listenkapazitt der JDialogs im Controller berschritten, das erste Element wurde wieder entfernt!");
		}
	}

	/**
	 * 
	 * Daten weiterreichen, gekapselt in ein ResultContent-Objekt
	 * 
	 * @param results
	 *            Liste mit eventuellen Ergebnissen
	 */
	private void tellChangedData(Class t, List<Object> results) {
		if (t != null) {
			submitData(new ResultContent(t, results));
		} else {
			submitData(null);
		}
	}

	/**
	 * benachrichtige Views, dass Daten gendert wurden.
	 * 
	 * @param object
	 *            gendertes Objekt
	 */
	private void submitData(Object object) {
		// setChanged() markiert das Objekt als sendebereit, sodass bei
		// notifyObservers() Meldungen gegeben werden.
		setChanged();
		notifyObservers(object);
	}
}
