package view;

import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.HeadlessException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.KeyEvent;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.List;
import java.util.Observable;
import java.util.Observer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.metamodel.Attribute;
import javax.persistence.metamodel.ManagedType;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;

import main.CustomItemListener;
import main.ProjektStatus;
import main.ResultContent;
import main.UserRole;

import controller.Controller;
import databaseModel.Ansprechpartner;
import databaseModel.Projekt;
import databaseModel.Student;
import databaseModel.Unternehmen;

/**
 * Basisklasse fr JDialogs
 * 
 * @author Thomas Kramer, eMail: th-rzv@gmx.de
 * 
 */
public abstract class AbstractDialog extends JDialog implements Observer {
	private static final long serialVersionUID = 1L;

	/**
	 * aktuelles Formular-Objekt
	 */
	protected JDialog activeDialog = null;

	/**
	 * Neu- oder ndernmodus des Formulars
	 */
	protected boolean newMode = true;

	/**
	 * vorgegebene Buttons
	 */
	private JButton okButton = null;
	private JButton closeButton = null;
	protected JPanel buttonPanel = null;

	/**
	 * Komponenten-List der Eingabefelder aus den abgeleiteten Klassen
	 */
	protected ArrayList<JComponent> mandatoryComponentList = new ArrayList<JComponent>();

	/**
	 * Objekt fr ndern-Modus
	 */
	protected Object changeObject = null;

	/**
	 * Daten-Listen
	 */
	protected List<Object> projekteListe = null;
	protected List<Object> unternehmenListe = null;
	protected List<Object> ansprechpartnerListe = null;
	protected List<Object> studentenListe = null;

	/**
	 * Controller-Objekt
	 */
	protected final Controller controller;
	protected final UserRole roleType;

	/**
	 * Default-Anzahl der mglichen Anzahl Zeichen in Textfeldern
	 */
	protected final int defaultInputFieldLength = 30;

	/**
	 * Default-Anzahl der angezeigten Spalten in Textfeldern
	 */
	protected final int defaultVisibleFieldLength = 30;

	/**
	 * Default-Anzahl der Zeilen in JTextArea-Feldern
	 */
	protected final int defaultTextAreaRows = 10;

	/**
	 * Default-Anzahl der Spalten in JTextArea-Feldern
	 */
	protected final int defaultTextAreaColumns = 30;

	/**
	 * Default-Farbe fr normale Felder
	 */
	protected final Color defaultColorField = null;

	/**
	 * Default-Farbe fr Pflichtfelder
	 */
	protected final Color defaultColorMandatoryField = Color.orange;

	/**
	 * Default-Farbe fr nichtausgefllte Pflichtfelder
	 */
	protected final Color defaultColorMissingMandatoryField = Color.red;

	/**
	 * Default-Tastenkombination fr bernehmen-Button
	 */
	protected final int defaultConfirmMnemonic = KeyEvent.VK_B;

	/**
	 * Default-Tastenkombination fr Abbrechen-Button
	 */
	protected final int defaultCancelMnemonic = KeyEvent.VK_C;

	/**
	 * Fehlermeldungsliste
	 */
	protected ArrayList<String> errorList = new ArrayList<String>();

	/**
	 * Konstruktor
	 * 
	 * @param controller
	 *            Controller
	 * @param label
	 *            Bezeichnung
	 */
	public AbstractDialog(JFrame owner, Controller controller, String label,
			UserRole roleType) {
		super(owner, label);
		this.controller = controller;
		this.roleType = roleType;
		// activeDialog kann nicht im Konstruktor gesetzt werden

		/**
		 * untere Buttonleiste, bernehmen und Abbrechen
		 */
		buttonPanel = new JPanel();
		buttonPanel.setLayout(new FlowLayout());
		okButton = new JButton("bernehmen");
		okButton.setMnemonic(defaultConfirmMnemonic);
		okButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						/**
						 * Default-Farben fr Eingabefelder setzen
						 */
						initializeFieldsWithColors();

						saveData();
					}
				});
			}
		});

		closeButton = new JButton("Abbrechen");
		closeButton.setMnemonic(defaultCancelMnemonic);
		closeButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						closeForm();
					}
				});
			}
		});
		buttonPanel.add(okButton);
		buttonPanel.add(closeButton);

		/**
		 * beim Drcken der Escape-Taste soll das aktuelle Fenster geschlossen
		 * werden
		 */
		getRootPane().registerKeyboardAction(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				closeForm();
			}
		}, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
				JComponent.WHEN_IN_FOCUSED_WINDOW);
	}

	/**
	 * Initialisieren der Eingabefelder mit den Default-Farben, sollte aus einem
	 * Thread heraus aufgerufen werden.
	 */
	protected void initializeFieldsWithColors() {
		for (JComponent c : mandatoryComponentList) {
			c.setBackground(defaultColorMandatoryField);
		}
	}

	/**
	 * bernehmen von neuen oder genderten Datenstzen, Methodensignatur sollte
	 * in abgeleiteten Klassen berschrieben werden
	 */
	protected void saveData() {
	}

	/**
	 * Auslesen der Feldlngen von Datenbank-Spalten aus den Entity-Klassen, die
	 * dazugehrigen Annotationen mssen hndisch gesetzt werden
	 * 
	 * @param classType
	 *            Klasse
	 * @param columnName
	 *            Spaltenname
	 * @return zulssige Anzahl Spalten zum Speichern
	 */
	protected int getColumnLength(Class classType, String columnName) {
		int columnLength = defaultInputFieldLength;

		try {
			Column column = classType.getDeclaredField(columnName)
					.getAnnotation(Column.class);
			if (column != null) {
				columnLength = column.length();
			}
		} catch (NoSuchFieldException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (SecurityException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return columnLength;
	}

	/**
	 * Vorselektion beim ffnen eines Fensters - das Formular sollte dazu direkt
	 * die update-Methode berschreiben, um die Daten zu empfangen. Das
	 * Observer-Pattern funktioniert aus dem Konstruktor heraus nicht, daher
	 * muss die update-Methode hier direkt aufgerufen werden.
	 * 
	 * @param classType
	 *            Klasse
	 */
	protected void preSelection(Class classType) {
		try {
			List<Object> objects = Controller.rmiServer.getData(classType);
			update(null, new ResultContent(classType, objects));
		} catch (RemoteException e2) {
			// TODO Auto-generated catch block
			e2.printStackTrace();
		}
	}

	/**
	 * Methodensignaturen fr abgeleitete Klassen zur Verfgung stellen, erspart
	 * einen expliziten Cast in der OpenDialog-Methode des MainFrames
	 * 
	 */
	public void initialize(final String label, final Object projekt) {
	}

	/**
	 * Zurckliefern der Daten ber das Observer-Pattern, sollte in den
	 * abgeleiteten Klassen berschrieben und tatschlich implementiert werden
	 * 
	 * In der Default-Version wird zumindest das kaskadierende Lschen von
	 * Detaildatenstzen abgefangen
	 */
	@Override
	public void update(Observable arg0, final Object arg1) {
		if (arg1 != null) {
			Class listType = ((ResultContent) arg1).getListType();
			List<Object> listContent = ((ResultContent) arg1).getListContent();

			if (changeObject != null) {
				if (listType == changeObject.getClass()) {
					closeFormOnDeleteCascading(changeObject.getClass(),
							listContent);
				}
			}
		}
	}

	/**
	 * berprfen der eingegebenen E-Mail-Adresse auf Gltigkeit
	 * 
	 * @param emailField
	 *            E-Mail-Eingabefeld
	 */
	protected void checkValidEmail(JTextField emailField) {
		if (!emailField.getText().trim().equals("")) {
			/**
			 * E-Mail-RegEx-Pattern, genommen von
			 * http://www.devx.com/tips/Tip/28334
			 */
			Pattern p = Pattern.compile(".+@.+\\.[a-z]+");
			Matcher m = p.matcher(emailField.getText().trim());

			if (!m.matches()) {
				errorList
						.add("Es wurde keine gltige E-Mail-Adresse eingegeben!");
				emailField.setBackground(defaultColorMissingMandatoryField);
			}
		}
	}

	/**
	 * berprfen der eingegebenen Matrikelnummer auf Gltigkeit
	 * 
	 * @param matField
	 *            Matrikelnummer-Eingabefeld
	 */
	protected void checkValidOnlyNumbers(JTextField matField) {
		if (!matField.getText().trim().equals("")) {
			/**
			 * Nur-Ziffern-RegEx-Pattern, genommen von
			 * http://docs.oracle.com/javase
			 * /1.4.2/docs/api/java/util/regex/Pattern.html
			 */
			Pattern p = Pattern.compile("\\p{Digit}+");
			Matcher m = p.matcher(matField.getText().trim());

			if (!m.matches()) {
				errorList
						.add("Das Matrikelnummernfeld darf nur Ziffern enthalten!");
				matField.setBackground(defaultColorMissingMandatoryField);
			}
		}
	}

	/**
	 * Updaten des Inhalts von Comboboxen und wiederherstellen der alten Auswahl
	 * 
	 * @param listType
	 *            Klassentyp der Objektliste vom zurckgelieferten
	 *            Observer-Pattern
	 * @param listContent
	 *            Objektliste vom zurckgelieferten Observer-Pattern
	 * @param newMode
	 *            Neu- oder ndern-Modus des Formulars
	 * @param comboBox
	 *            JComboBox, die upgedatetet werden soll
	 * @param oldSelectedData
	 *            zuvor selektiertes Objekt
	 * @param specialCase
	 *            Sonderfall fr den 2. und 3. Studenten eines Projekts, diese
	 *            sind optional und mssen als ersten Eintrag eine Leerzeile
	 *            erhalten
	 */
	protected void updateGUIData(Class listType, List<Object> listContent,
			boolean newMode, JComboBox comboBox, Object oldSelectedData,
			boolean specialCase) {
		/**
		 * whrend des Hinzufgen/Lschens von Eintrgen aus der ComboBox soll
		 * nicht das ItemStateChanged-Event ausgelst werden, daher wird die
		 * ComboBox auf inaktiv geschaltet.
		 * 
		 * Voraussetzungen: Jede ComboBox hat nur genau EINEN ItemListener, und
		 * dieser muss vom Typ CustomItemListener sein
		 */
		((CustomItemListener) comboBox.getItemListeners()[0]).setActive(false);

		/**
		 * zurcksetzen
		 */
		Integer comboboxIndex = 0;
		boolean found = false;
		comboBox.setSelectedIndex(-1);
		comboBox.removeAllItems();

		/**
		 * Combobox fllen
		 */
		if (specialCase) {
			comboBox.addItem("");
			comboboxIndex = 1;
		}
		Integer foundIndex = comboboxIndex;

		/**
		 * a ist aus der neu ausgelesenen Liste, b aus der alten Liste
		 */
		for (Object a : listContent) {
			try {
				try {
					/**
					 * Combobox fllen mit Inhalt aus der neuen Liste
					 */
					Class classA = a.getClass();
					Method methodClassAGetName = classA.getMethod("getName");

					String resultClassAGetName = (String) methodClassAGetName
							.invoke(a);
					comboBox.addItem(resultClassAGetName);

					/**
					 * ndern-Modus: Wenn IDs bereinstimmen vorherige Auswahl
					 * erneut setzen
					 */
					if (!newMode) {
						if (oldSelectedData != null) {
							Method methodClassAGetId = classA
									.getMethod("getId");

							Class classB = oldSelectedData.getClass();
							Method methodClassBGetId = classB
									.getMethod("getId");

							int resultClassAGetId = (Integer) methodClassAGetId
									.invoke(a);
							int resultClassBGetId = (Integer) methodClassBGetId
									.invoke(oldSelectedData);

							if (resultClassAGetId == resultClassBGetId) {
								foundIndex = comboboxIndex;
								found = true;
							}
						}
					}
				} catch (IllegalAccessException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (IllegalArgumentException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (InvocationTargetException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			} catch (NoSuchMethodException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (SecurityException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			comboboxIndex++;
		}

		/**
		 * setze Vorselektion
		 */
		if (newMode) {
			if (listContent.size() > 0) {
				comboBox.setSelectedIndex(0);
			}
		} else {
			if (found) {
				/**
				 * neue Selektion setzen
				 */
				comboBox.setSelectedIndex(foundIndex);
			}
		}

		/**
		 * aktiviere wieder ItemListener
		 */
		((CustomItemListener) comboBox.getItemListeners()[0]).setActive(true);

		/**
		 * hndisches Feuern des ItemChanged-Events sollte eigentlich reichen,
		 * ist aber leider zu unzuverlssig in Java.
		 * 
		 * Ausfhrung dessen, was mit der Vorselektion geschehen soll, muss
		 * daher hndisch nach der Vorselektion geschehen
		 */
		/*
		 * Object selectedItem = comboBox.getSelectedItem();
		 * ((CustomItemListener) comboBox.getItemListeners()[0])
		 * .itemStateChanged(new ItemEvent(comboBox,
		 * ItemEvent.ITEM_STATE_CHANGED, selectedItem, ItemEvent.SELECTED));
		 */

		/**
		 * Daten in interne Listen bernehmen, sehr wichtig damit sie
		 * bereinstimmen mit den Comboboxen
		 */
		if (listType == Ansprechpartner.class) {
			ansprechpartnerListe = listContent;
		} else if (listType == Projekt.class) {
			projekteListe = listContent;
		} else if (listType == Unternehmen.class) {
			unternehmenListe = listContent;
		} else if (listType == Student.class) {
			studentenListe = listContent;
		}
	}

	/**
	 * berprfen ob Datensatz bereits vorhanden ist
	 * 
	 * @param classType
	 *            Klasse
	 * @param checkingField
	 *            zu berprfendes Feld
	 * @param checkingClauseComponent
	 *            Text, fr das zu berprfende Feld
	 */
	protected void checkExistingRecord(Class classType, String checkingField,
			Object checkingClauseComponent) {
		boolean OK = true;
		String checkingClause = "";

		if (checkingClauseComponent.getClass() == JTextField.class) {
			checkingClause = ((JTextField) checkingClauseComponent).getText()
					.trim();
		}

		if (!checkingClause.equals("")) {
			try {
				List<Object> existingRecords = Controller.rmiServer
						.getDataWithClause(classType, checkingField,
								checkingClause);

				if (newMode) {
					OK = !(existingRecords.size() > 0);
				} else {
					/**
					 * Im ndern-Modus muss das eigene Projekt ausgeschlossen
					 * werden
					 */
					for (Object a : existingRecords) {
						Class classA = a.getClass();
						Class classB = changeObject.getClass();
						try {
							Method methodClassA = classA.getMethod("getId");
							Method methodClassB = classB.getMethod("getId");

							try {
								int resultClassA = (Integer) methodClassA
										.invoke(a);
								int resultClassB = (Integer) methodClassB
										.invoke(changeObject);

								if (resultClassA != resultClassB) {
									OK = false;
									break;
								}
							} catch (IllegalAccessException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							} catch (IllegalArgumentException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							} catch (InvocationTargetException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}

						} catch (NoSuchMethodException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						} catch (SecurityException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
					}
				}
			} catch (RemoteException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

			if (!OK) {
				errorList.add("Dieser Datensatz ist bereits vorhanden!");
			}
		}

		if (!OK) {
			if (checkingClauseComponent.getClass() == JTextField.class) {
				((JTextField) checkingClauseComponent)
						.setBackground(defaultColorMissingMandatoryField);
			}
		}
	}

	/**
	 * Methode um das aktive Formular, welches sich im ndern-Modus befindet,
	 * automatisch zu schlieen wenn ein referenziertes Objekt einer anderen
	 * Datenbank-Tabelle gelscht wurde.
	 * 
	 * Stichpunkt "Kaskadierendes Lschen von Detaildatenstzen" - man kann die
	 * Datenbank so konfigurieren dass sie beim Lschen eines Datensatzes
	 * automatisch referenzierte Datenstze anderer Tabellen mitlscht.
	 * 
	 * Es wird Reflection benutzt, da jede JPA-Entity eine getId-Methode besitzt
	 * - auf diese Weise wird vermieden dass diese Methode spter hndisch
	 * erweitert werden muss
	 * 
	 * @param classType
	 *            Klasse
	 * @param listContent
	 *            Liste von Objekten
	 */
	protected void closeFormOnDeleteCascading(Class classType,
			List<Object> listContent) {
		if (!newMode && changeObject != null && activeDialog.isVisible()) {
			boolean existingFurthermore = false;

			for (Object a : listContent) {
				Class classA = a.getClass();
				Class classB = changeObject.getClass();
				try {
					Method methodClassA = classA.getMethod("getId");
					Method methodClassB = classB.getMethod("getId");

					try {
						int resultClassA = (Integer) methodClassA.invoke(a);
						int resultClassB = (Integer) methodClassB
								.invoke(changeObject);

						if (resultClassA == resultClassB) {
							existingFurthermore = true;
							break;
						}
					} catch (IllegalAccessException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					} catch (IllegalArgumentException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					} catch (InvocationTargetException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}

				} catch (NoSuchMethodException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (SecurityException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}

			if (!existingFurthermore) {
				errorList
						.add("Der aktuell bearbeitete Datensatz wurde direkt oder einer seiner referenzierten Datenstze gelscht, wodurch er automatisch mitgelscht wurde!");
				errorList
						.add("Stichwort \"kaskadierendes Lschen von Detaildatenstzen!\"!");
				errorList
						.add("Das aktive Fenster wird automatisch geschlossen!");

				showErrors();
				closeForm();
			}
		}
	}

	/**
	 * Zurckliefern der gesammelten Fehlermeldungen aus der errorList als einen
	 * zusammenhngenden String - wenn man direkt .toString() auf die ArrayList
	 * ausfhrt erhlt man eckige Klammern davor und danach sowie Trennung durch
	 * Kommas und alles in einer Zeile.
	 * 
	 * @return zusammengehngten String
	 */
	private String getErrorString() {
		String result = "";
		for (String s : errorList) {
			result += s;
			result += System.getProperty("line.separator");
		}
		return result;
	}

	/**
	 * Zeige Fehlermeldungen an und leere Fehlerliste anschlieend
	 */
	protected void showErrors() {
		if (errorList.size() > 0) {
			JOptionPane.showMessageDialog(null, getErrorString(), "Fehler",
					JOptionPane.ERROR_MESSAGE);

			/**
			 * Ein Lschen der Fehlerliste hier erspart ein separates Lschen
			 */
			errorList.clear();
		}
	}

	/**
	 * zentrale Schlieen-Methode des Formulars
	 */
	protected void closeForm() {
		/**
		 * Formular schlieen, vor dem Schlieen auf Neu-Modus zurcksetzen
		 */
		newMode = true;
		dispose();
	}
}
